@@ -530,9 +530,9 @@ static void
 _cairo_xcb_surface_get_font_options (void *abstract_surface,
 				     cairo_font_options_t *options)
 {
-    /* XXX  copy from xlib */
-    _cairo_font_options_init_default (options);
-    _cairo_font_options_set_round_glyph_positions (options, CAIRO_ROUND_GLYPH_POS_ON);
+    cairo_xcb_surface_t *surface = abstract_surface;
+
+    *options = *_cairo_xcb_screen_get_font_options (surface->screen);
 }
 
 static cairo_status_t
@@ -646,9 +646,8 @@ _put_shm_image_boxes (cairo_xcb_surface_t    *surface,
 						     shm_info->offset);
 	    }
 	}
+	return CAIRO_INT_STATUS_SUCCESS;
     }
-
-    return CAIRO_INT_STATUS_SUCCESS;
 #endif
 
     return CAIRO_INT_STATUS_UNSUPPORTED;
@@ -692,15 +691,15 @@ _put_image_boxes (cairo_xcb_surface_t    *surface,
 			    int y = _cairo_fixed_integer_part (b->p1.y);
 			    int width = _cairo_fixed_integer_part (b->p2.x - b->p1.x);
 			    int height = _cairo_fixed_integer_part (b->p2.y - b->p1.y);
-			    _cairo_xcb_connection_put_image (surface->connection,
-							     surface->drawable, gc,
-							     width, height,
-							     x, y,
-							     image->depth,
-							     image->stride,
-							     image->data +
-							     x * PIXMAN_FORMAT_BPP (image->pixman_format) / 8 +
-							     y * image->stride);
+			    _cairo_xcb_connection_put_subimage (surface->connection,
+								surface->drawable, gc,
+								x, y,
+								width, height,
+								PIXMAN_FORMAT_BPP (image->pixman_format) / 8,
+								image->stride,
+								x, y,
+								image->depth,
+								image->data);
 
 		    }
 	    }
@@ -1432,7 +1431,7 @@ cairo_xcb_surface_set_size (cairo_surface_t *abstract_surface,
     }
 
 
-    if (abstract_surface->type != CAIRO_SURFACE_TYPE_XCB) {
+    if ( !_cairo_surface_is_xcb(abstract_surface)) {
 	_cairo_surface_set_error (abstract_surface,
 				  _cairo_error (CAIRO_STATUS_SURFACE_TYPE_MISMATCH));
 	return;
@@ -1486,7 +1485,7 @@ cairo_xcb_surface_set_drawable (cairo_surface_t *abstract_surface,
     }
 
 
-    if (abstract_surface->type != CAIRO_SURFACE_TYPE_XCB) {
+    if ( !_cairo_surface_is_xcb(abstract_surface)) {
 	_cairo_surface_set_error (abstract_surface,
 				  _cairo_error (CAIRO_STATUS_SURFACE_TYPE_MISMATCH));
 	return;
